/*
 * Copyright (c) 2007 Sergey Matveychuk
 *      Yandex, LLC.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of the company nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/select.h>
#include <net/ethernet.h>
#include <net/if.h>
#include <net/if_dl.h>
#include <netdb.h>
#include <ifaddrs.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <sys/ioctl.h>
#include <sys/mac.h>
#include <pcap.h>
#include <syslog.h>
#include <errno.h>
#include <pthread.h>
#include <sys/queue.h>
#include <libutil.h>

#include "dhcprelya.h"

#define PURGE_INTERVAL	600	/* sec */

static int debug=0;

struct pidfh *pfh=NULL;
int bootps_port, bootpc_port;
struct interface *ifs[IF_MAX];
struct server *servers[SERVER_MAX];
int if_num=0;	/* interfaces number */
int srv_num=0;	/* servers number */
int queue_size=0;
int timer_val;	/* own timer */
pthread_mutex_t queue_lock, request_lock;

void
usage(char *prgname)
{
    fprintf(stderr, "Usage: %s [-d] [-p<pidfile>] -i <ifname>... <dhcp_server>...\n", prgname);
    exit(EX_OK);
}

void
process_error(int ret_code, char *fmt, ...)
{
    va_list ap;
    char buf[1024];

    va_start(ap, fmt);

    if(pfh != NULL)
	pidfile_remove(pfh);

    vsprintf(buf, fmt, ap);
    syslog(LOG_ERR, buf);
    errx(ret_code, buf);

    /* does not reach */
    va_end(ap);
}

char *
print_mac(u_char *s, u_char *buf)
{
    sprintf(buf, "%02x:%02x:%02x:%02x:%02x:%02x",
	    s[0],s[1],s[2],s[3],s[4],s[5]);
    return buf;
}

char *
print_ip(ip_addr_t ip, u_char *buf)
{
    union {
	ip_addr_t	n;
	u_char		b[4];
    } u;

    u.n = ip;
    sprintf(buf, "%u.%u.%u.%u", u.b[0], u.b[1], u.b[2], u.b[3]);
    return buf;
}

char *
print_xid(u_int32_t ip, u_char *buf)
{
    union {
	u_int32_t	n;
	u_char		b[4];
    } u;

    u.n = ip;
    sprintf(buf, "0x%02x%02x%02x%02x", u.b[0], u.b[1], u.b[2], u.b[3]);
    return buf;
}

void
get_mac(char *if_name, char *if_mac)
{
    struct ifaddrs *ifaphead;
    int found=0;
    struct ifaddrs *ifap;
    struct sockaddr_dl *sdl=NULL;

    if(getifaddrs(&ifaphead) != 0)
	process_error(EX_RES, "getifaddrs");

    for(ifap = ifaphead; ifap && !found; ifap = ifap->ifa_next)
	if (strcmp(ifap->ifa_name, if_name) == 0) {
	    found = 1;
	    sdl = (struct sockaddr_dl *)ifap->ifa_addr;
	    if(sdl)
		memcpy(if_mac, LLADDR(sdl), sdl->sdl_alen);
	}

    freeifaddrs(ifaphead);
    if(!found)
	process_error(EX_RES, "can't find mac for interface %s", if_name);
}

void get_ip(char *iname, ip_addr_t *ip, ip_addr_t *mask)
{
    int s;
    struct ifreq ifr;
    struct sockaddr_in *saddr;

    if((s = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
	process_error(EX_RES, "get_ip: socket");

    strcpy(ifr.ifr_name, iname);
    if (ioctl(s, SIOCGIFADDR, &ifr) < 0)
	process_error(EX_RES, "Can't get IP for %s", iname);

    saddr = (struct sockaddr_in*)&ifr.ifr_addr;
    memcpy(ip, &saddr->sin_addr.s_addr, sizeof(ip_addr_t));

    if (ioctl(s, SIOCGIFNETMASK, &ifr) < 0)
	process_error(EX_RES, "Can't get netmask for %s", iname);
    memcpy(mask, &saddr->sin_addr.s_addr, sizeof(ip_addr_t));

    close(s);
}

void
add_request(u_int32_t xid, ip_addr_t ip, int if_idx)
{
    struct request *req;
    char buf[32];

    req = malloc(sizeof(struct request));

    req->xid = xid;
    if(ip == INADDR_BROADCAST)
	req->broadcast = 1;
    else
	req->broadcast = 0;
    req->if_idx = if_idx;
    req->timestamp = time(NULL);

    pthread_mutex_lock(&request_lock);
    SLIST_INSERT_HEAD(&req_head, req, entries);
    pthread_mutex_unlock(&request_lock);
}

struct request *
find_request(u_int32_t xid)
{
    struct request *req;

    pthread_mutex_lock(&request_lock);
    SLIST_FOREACH(req, &req_head, entries)
	if(req->xid == xid)
	    return req;
    pthread_mutex_unlock(&request_lock);

    return NULL;
}

void
purge_requests()
{
    int n=0;
    struct request *req, *req_temp;
    time_t now;

    time(&now);
    /* Remove entries older than PURGE_INTERVAL */
    pthread_mutex_lock(&request_lock);
    SLIST_FOREACH_SAFE(req, &req_head, entries, req_temp)
	if(now-req->timestamp > 600) {
	    SLIST_REMOVE(&req_head, req, request, entries);
	    free(req);
	    n++;
	}
    pthread_mutex_unlock(&request_lock);

    if(n && debug)
	syslog(LOG_WARNING, "Requests purged: %d", n);
}

void
open_interface(char *iname)
{
    int j, x=1;
    struct ifreq ifr;
    struct bpf_program fp;
    struct sockaddr_in baddr;
    char errbuf[PCAP_ERRBUF_SIZE], file[32], buf[256], logbuf[256];
    char *p;

    if(if_num >= IF_MAX-1)
	process_error(EX_RES, "too many interfaces");

    ifs[if_num] = malloc(sizeof(struct interface));
    if(ifs[if_num] == NULL)
	process_error(EX_MEM, "malloc");

    ifs[if_num]->idx = if_num;
    strncpy(ifs[if_num]->name, iname, INTF_NAME_LEN);

    get_mac(iname, ifs[if_num]->mac);

    /* Looking for a free BPF device and open it */
    for (j = 0; j < 255; j++) {
	snprintf(file, sizeof(file), "/dev/bpf%d", j);
	ifs[if_num]->bpf = open(file, O_WRONLY);
	if(ifs[if_num]->bpf != -1 || errno != EBUSY)
	    break;
    }
    /* Bind BPF to an interface */
    bzero(&ifr, sizeof(ifr));
    strlcpy(ifr.ifr_name, iname, sizeof(ifr.ifr_name));
    if(ioctl(ifs[if_num]->bpf, BIOCSETIF, (char *)&ifr) < 0)
	process_error(EX_RES, "Can't BIOCSETIF");

    if((ifs[if_num]->cap = pcap_open_live(iname, 1500, 1, 100, errbuf)) == NULL)
	process_error(EX_RES, "pcap_open_live(%s): %s", iname, errbuf);
    if(pcap_compile(ifs[if_num]->cap, &fp, "udp and (port bootpc or port bootps)", 0, 0) < 0)
	process_error(EX_RES, "pcap_compile");
    if(pcap_setfilter(ifs[if_num]->cap, &fp) < 0)
	process_error(EX_RES, "pcap_setfilter");

    if((ifs[if_num]->fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
	process_error(EX_RES, "socket for listener at %s", iname);
    if(setsockopt(ifs[if_num]->fd, SOL_SOCKET, SO_BROADCAST, (char *)&x, sizeof(x)) < 0)
	process_error(EX_RES, "setsockopt: SO_BROADCAST");
    if(setsockopt(ifs[if_num]->fd, SOL_SOCKET, SO_REUSEADDR, (char *)&x, sizeof(x)) < 0)
	process_error(EX_RES, "setsockopt: SO_REUSEADDR");

    get_ip(iname, &ifs[if_num]->ip, &ifs[if_num]->mask);

    bzero(&baddr, sizeof(baddr));
    baddr.sin_family = AF_INET;
    baddr.sin_port = bootps_port;
    memcpy(&baddr.sin_addr.s_addr, &ifs[if_num]->ip, sizeof(ip_addr_t));
    if(bind(ifs[if_num]->fd, (struct sockaddr*)&baddr, sizeof(baddr)) < 0)
	process_error(EX_RES, "bind");

    sprintf(logbuf, "Listen at %s: %s/%s, %s", iname, print_ip(ifs[if_num]->ip, buf), print_ip(ifs[if_num]->mask, buf+16), print_mac(ifs[if_num]->mac, buf+32));
    if(debug)
	puts(logbuf);
    syslog(LOG_WARNING, logbuf);

    if_num++;
}

void
open_server(char *name)
{
    struct hostent *hostent;
    char buf[16], logbuf[256];

    if(srv_num >= SERVER_MAX-1)
	process_error(EX_RES, "too many servers");

    servers[srv_num] = malloc(sizeof(struct server));
    if(servers[srv_num] == NULL)
	process_error(EX_MEM, "malloc");

    servers[srv_num]->name = malloc(strlen(name));
    if(servers[srv_num]->name == NULL)
	process_error(EX_MEM, "malloc");
    strcpy(servers[srv_num]->name, name);

    if((hostent = gethostbyname(name)) == NULL)
	process_error(EX_NOHOST, "Host \"%s\" not found", name);

    bzero(&servers[srv_num]->sockaddr, sizeof(struct sockaddr_in));
    servers[srv_num]->sockaddr.sin_family = AF_INET;
    servers[srv_num]->sockaddr.sin_port = bootps_port;
    memcpy(&(servers[srv_num]->sockaddr.sin_addr.s_addr), hostent->h_addr, hostent->h_length);

    sprintf(logbuf, "DHCP server #%d: %s (%s)", srv_num+1, name, print_ip(servers[srv_num]->sockaddr.sin_addr.s_addr, buf));

    srv_num++;

    if(debug)
	puts(logbuf);
    syslog(LOG_WARNING, logbuf);
}

void *
timer_thread(void *param)
{
    while(1) {
	sleep(1);
	timer_val++;
    }
}

void *
listener(void *param)
{
    int i,n,ignore=0;
    struct interface *intf = param;
    struct pcap_pkthdr *header;
    const u_char *packet;
    u_char *p;
    struct queue *q;
    struct tm tm;
    char buf[18*2], logbuf[256];

    while(1) {
	if((n=pcap_next_ex(intf->cap, &header, &packet)) > 0) {
	    /* Ignore replay from server, we've sent it */
	    if(((struct dhcp_packet*)(packet+ETHER_HDR_LEN+DHCP_UDP_OVERHEAD))->op == BOOTREPLY)
		continue;
	    /* If a packet with our MAC, it's our one. Ignore it */
	    for(i = 0; i < if_num; i++) {
		if(memcmp(((struct ether_header*)packet)->ether_shost,
			    ifs[i]->mac, ETHER_ADDR_LEN) == 0)
			ignore = 1;
	    }

	    if(ignore) {
		ignore = 0;
		continue;
	    }

	    if(debug) {
		localtime_r((const time_t*)&header->ts.tv_sec, &tm);
		sprintf(logbuf, "%02d:%02d:%02d.%06lu %s got %d/%d: %s -> %s",\
			tm.tm_hour, tm.tm_min, tm.tm_sec, header->ts.tv_usec,
			intf->name, header->caplen, header->len,
			print_mac(((struct ether_header*)packet)->ether_shost, buf),
			print_mac(((struct ether_header*)packet)->ether_dhost, buf+18)
			);
		puts(logbuf);
	    }

	    q = malloc(sizeof(struct queue));
	    p = malloc(header->caplen);
	    memcpy(p, packet, header->caplen);
	    q->packet = p;
	    q->len = header->caplen;
	    q->if_idx = intf->idx;
	    pthread_mutex_lock(&queue_lock);
	    STAILQ_INSERT_TAIL(&q_head, q, entries);
	    queue_size++;
	    pthread_mutex_unlock(&queue_lock);
	}
    }
}

void
process_server_answer()
{
    struct request *req;
    struct dhcp_packet *dhcp;
    struct ether_header *eh;
    struct ip *ip;
    struct udphdr *udp;
    char buf[1500], packet[1500];
    char pbuf[11+16+19], logbuf[256];
    struct sockaddr_in from_addr;
    socklen_t from_len=sizeof(from_addr);
    int psize, i, fdmax=0;
    fd_set fds;
    struct timeval timeout, tv;
    struct tm tm;
    struct timezone tz;

    dhcp = (struct dhcp_packet *)&buf;
    FD_ZERO(&fds);
    for(i = 0; i < if_num; i++) {
	FD_SET(ifs[i]->fd, &fds);
	if(fdmax < ifs[i]->fd)
	    fdmax = ifs[i]->fd;
    }

    timeout.tv_sec = 0;
    timeout.tv_usec = 10;
    if(select(fdmax+1, &fds, NULL, NULL, &timeout)) {
	for(i = 0; i < if_num; i++)
	    if(FD_ISSET(ifs[i]->fd, &fds)) {
		bzero(&from_addr, sizeof(from_addr));
		/*from_addr.sin_family=AF_INET;
		from_addr.sin_port = bootps_port;*/
    		from_len=sizeof(from_addr);

		psize = recvfrom(ifs[i]->fd, buf, 1500, 0, (struct sockaddr*)&from_addr, &from_len);
		if(debug) {
		    gettimeofday(&tv, &tz);
		    localtime_r((const time_t*)&tv.tv_sec, &tm);
		    sprintf(logbuf, "%02d:%02d:%02d.%06lu got from server (%s) XID: %s for %s",
			    tm.tm_hour, tm.tm_min, tm.tm_sec, tv.tv_usec,
			    print_ip(from_addr.sin_addr.s_addr, pbuf),
			    print_xid(dhcp->xid, pbuf+16),
			    print_mac(dhcp->chaddr, pbuf+27)
			    );
		    puts(logbuf);
		}
	    }
    } else {
	return;
    }

    req = find_request(dhcp->xid);
    if(req == NULL) {
	sprintf(logbuf, "XID %s not found!", print_xid(dhcp->xid, pbuf));
	if(debug)
	    puts(logbuf);
	syslog(LOG_ERR, logbuf);
	return;
    }
    /* Generate a new packet to send to a client */
    eh = (struct ether_header *)packet;
    ip = (struct ip *)(packet+ETHER_HDR_LEN);
    udp = (struct udphdr *)(packet+ETHER_HDR_LEN+sizeof(struct ip));

    memcpy(eh->ether_dhost, dhcp->chaddr, ETHER_ADDR_LEN);
    memcpy(eh->ether_shost, ifs[req->if_idx]->mac, ETHER_ADDR_LEN);
    eh->ether_type = htons(ETHERTYPE_IP);

    ip->ip_v = IPVERSION;
    ip->ip_hl = 5;		/* IP header length is 5 byte (no options) */
    ip->ip_tos = IPTOS_LOWDELAY;
    ip->ip_len = htons(sizeof(struct ip)+sizeof(struct udphdr)+psize);
    ip->ip_id = 0;
    ip->ip_off = 0;
    ip->ip_ttl = 16;
    ip->ip_p = IPPROTO_UDP;
    ip->ip_sum = 0;
    memcpy(&ip->ip_src, &ifs[req->if_idx]->ip, sizeof(ip_addr_t));
    if(req->broadcast)
	ip->ip_dst.s_addr = INADDR_BROADCAST;
    else
	memcpy(&ip->ip_dst, &dhcp->yiaddr, sizeof(ip_addr_t));

    udp->uh_sport = bootps_port;
    udp->uh_dport = bootpc_port;
    udp->uh_ulen = htons(sizeof(struct udphdr)+psize);
    udp->uh_sum = 0;		/* UDP checksum is optional. we don't care it. */

    memcpy(packet+ETHER_HDR_LEN+DHCP_UDP_OVERHEAD, buf, psize);
    ip->ip_sum = htons(ip_checksum((const char*)ip, sizeof(struct ip)));

    write(ifs[req->if_idx]->bpf, packet, ETHER_HDR_LEN+DHCP_UDP_OVERHEAD+psize);
    if(debug) {
    gettimeofday(&tv, &tz);
	localtime_r((const time_t*)&tv.tv_sec, &tm);
	sprintf(logbuf, "%02d:%02d:%02d.%06lu sent to %s XID: %s for %s (%s)",
		tm.tm_hour, tm.tm_min, tm.tm_sec, tv.tv_usec,
		print_ip(ip->ip_dst.s_addr, pbuf),
		print_xid(dhcp->xid, pbuf+16), print_mac(dhcp->chaddr, pbuf+27),
		ifs[req->if_idx]->name);
	puts(logbuf);
    }
}

int
sanity_check(struct queue *q)
{
    struct ether_header *eh;
    struct ip *ip;

    eh = (struct ether_header *)q->packet;
    ip = (struct ip *)(q->packet+ETHER_HDR_LEN);

    if(q->len < ETHER_HDR_LEN+DHCP_FIXED_LEN) {
	syslog(LOG_ERR, "length too little -- packet ignore");
	if(debug)
	    printf("length too little -- packet ignore\n");
	return 0;
    }
    if(eh->ether_type != 8) {
	syslog(LOG_ERR, "wrong ether type -- packet ignore\n");
	if(debug)
	    printf("wrong ether type -- packet ignore\n");
	return 0;
    }
    if(ip->ip_p != IPPROTO_UDP) {
	syslog(LOG_ERR, "not an udp packet -- packet ignore\n");
	if(debug)
	    printf("not an udp packet -- packet ignore\n");
	return 0;
    }

    return 1;
}

/* Get one packet from queue and process it */
void
process_queue()
{
    int i;
    struct queue *q;
    struct dhcp_packet *dhcp;
    struct tm tm;
    struct timeval tv;
    struct timezone tz;
    char buf[11+27], logbuf[256];

    pthread_mutex_lock(&queue_lock);
    q = STAILQ_FIRST(&q_head);
    STAILQ_REMOVE_HEAD(&q_head, entries);
    queue_size--;
    pthread_mutex_unlock(&queue_lock);

    if( sanity_check(q) ) {
	dhcp = (struct dhcp_packet *)(q->packet+ETHER_HDR_LEN+DHCP_UDP_OVERHEAD);

	dhcp->hops++;
	/* Check for a loop and ignore if so */
	if(dhcp->hops > 10) {
	    free(q->packet);
	    free(q);
	    return;
	}
	memcpy(&dhcp->giaddr, &ifs[q->if_idx]->ip, sizeof(ip_addr_t));
	for(i=0; i < srv_num; i++)
	    sendto(ifs[q->if_idx]->fd, dhcp, q->len-ETHER_HDR_LEN+DHCP_UDP_OVERHEAD,
		0, (struct sockaddr *)&servers[i]->sockaddr, sizeof(struct sockaddr_in));

	add_request(dhcp->xid,
			((struct ip*)(q->packet+ETHER_HDR_LEN))->ip_dst.s_addr,
			q->if_idx);

	if(debug) {
	    gettimeofday(&tv, &tz);
	    localtime_r((const time_t*)&tv.tv_sec, &tm);
	    sprintf(logbuf, "%02d:%02d:%02d.%06lu sent to %d server(s) XID: %s for %s",
		    tm.tm_hour, tm.tm_min, tm.tm_sec, tv.tv_usec,
		    srv_num,
		    print_xid(dhcp->xid, buf), print_mac(dhcp->chaddr, buf+11)
		    );
	    puts(logbuf);
	}
    }
    free(q->packet);
    free(q);
}

int
main(int argc, char *argv[])
{
    int c, i;
    pid_t opid;
    char prgname[80], filename[256], *p;
    char logbuf[256];
    struct servent *servent;
    pthread_t tid;
    time_t last_purge, now;

    if((servent = getservbyname("bootps", 0)) == NULL)
	errx(EX_UNAVAILABLE, "getservbyname(bootps)");
    bootps_port=servent->s_port;
    if((servent = getservbyname("bootpc", 0)) == NULL)
	errx(EX_UNAVAILABLE, "getservbyname(bootpc)");
    bootpc_port=servent->s_port;

    strncpy(prgname, argv[0], sizeof(prgname));
    filename[0] = '\0';
    while((c = getopt(argc, argv, "di:p:")) != -1) {
	switch(c) {
	    case 'd':
		debug++;
		break;
	    case 'i':
		open_interface(optarg);
		break;
	    case 'p':
		strncpy(filename, optarg, sizeof(filename));
		break;
	    default:
		usage(prgname);
	}
    }

    argc -= optind;
    argv += optind;

    if(optind == 0)
	argc--;

    if(argc < 1)
	usage(prgname);

    for(i=0; i<argc; i++)
	open_server(argv[i]);

    openlog("dhcprelya", LOG_NDELAY|LOG_PID, LOG_DAEMON);

    sprintf(logbuf, "Total interfaces: %d", if_num);
    if(debug)
	puts(logbuf);
    syslog(LOG_WARNING, logbuf);

    if(filename[0] == '\0') {
	strcpy(filename, "/var/run/");
	p = strrchr(prgname, '/');
	if(p == NULL)
	    p = prgname;
	else
	    p++;
	strcat(filename, p);
	strcat(filename, ".pid");
    }

    if((pfh = pidfile_open(filename, 0644, &opid)) == NULL) {
	if (errno == EEXIST)
	    errx(1, "Already run with PID %U. Exiting.", opid);
	errx(1, "Can't create PID file");
    }

    if(!debug)
	if(daemon(1,1) == -1)
	    process_error(1, "Can't daemonize. Exiting.");

    pidfile_write(pfh);

    syslog(LOG_WARNING, "Runned as %d", getpid());

    STAILQ_INIT(&q_head);
    SLIST_INIT(&req_head);

    pthread_mutex_init(&queue_lock, NULL);
    pthread_mutex_init(&request_lock, NULL);

    /* Run own timer */
    timer_val = 0;
    pthread_create(&tid, NULL, timer_thread, NULL);
    pthread_detach(tid);

    /* Create listeners for every interface */
    for(i = 0; i < if_num; i++) {
	pthread_create(&tid, NULL, listener, ifs[i]);
	pthread_detach(tid);
    }

    last_purge = timer_val;
    /* Main loop */
    while(1) {
	if(queue_size > 0)
	    process_queue();

	process_server_answer();

	now = timer_val;
	if(now-last_purge > PURGE_INTERVAL) {
	    purge_requests();
	    last_purge = now;
	}
    }
}
